import rebound
import reboundx
import unittest
import os
import numpy as np

class TestTides(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1., r=0.005)
        self.sim.add(m=1.e-3, a=0.05, e=0.1, r=0.0005)
        self.sim.move_to_com()
        self.rebx = reboundx.Extras(self.sim)
        self.force = self.rebx.load_force("tides_constant_time_lag")
        self.rebx.add_force(self.force)

    def test_conservation_star_highmratio(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1., r=0.005)
        self.sim.add(m=1., a=0.2, e=0.1, r=0.005)
        self.sim.move_to_com()
        self.rebx = reboundx.Extras(self.sim)
        self.force = self.rebx.load_force("tides_constant_time_lag")
        self.rebx.add_force(self.force)
        ps = self.sim.particles
        ps[0].params['tctl_k1'] = 0.04
        self.do_test_conservation()

    def test_conservation_planet_highmratio(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1., r=0.005)
        self.sim.add(m=1., a=0.2, e=0.1, r=0.005)
        self.sim.move_to_com()
        self.rebx = reboundx.Extras(self.sim)
        self.force = self.rebx.load_force("tides_constant_time_lag")
        self.rebx.add_force(self.force)
        ps = self.sim.particles
        ps[1].params['tctl_k1'] = 0.04
        self.do_test_conservation()
    
    def test_conservation_star(self):
        ps = self.sim.particles
        ps[0].params['tctl_k1'] = 0.04
        self.do_test_conservation()

    def test_conservation_planet(self):
        ps = self.sim.particles
        ps[1].params['tctl_k1'] = 0.4
        self.do_test_conservation()
    
    def test_conservation_star_movecom(self):
        # if you go much larger, IAS15 starts giving errors due to roundoff
        self.sim.particles[0].vy += 0.01 
        self.sim.particles[1].vy += 0.01 
        self.test_conservation_star()
    
    def test_conservation_planet_movecom(self):
        self.sim.particles[0].vy += 0.01 
        self.sim.particles[1].vy += 0.01 
        self.test_conservation_planet()

    def do_test_conservation(self):
        H0 = self.sim.calculate_energy() + self.rebx.tides_constant_time_lag_potential(self.force)
        self.sim.integrate(1.e3*self.sim.particles[1].P)
        H = self.sim.calculate_energy() + self.rebx.tides_constant_time_lag_potential(self.force)
        self.assertLess(abs((H-H0)/H0), 1.e-12)
        self.assertGreater(self.sim.particles[1].pomega, 1.e-6)

class TestTidesAnalytic(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=0.86, r = 0.78)   
        self.sim.add(m=3.e-6, a=1., e=0.05)
        self.sim.move_to_com()
        ps = self.sim.particles

        self.rebx = reboundx.Extras(self.sim)
        self.tides = self.rebx.load_force("tides_constant_time_lag")
        self.rebx.add_force(self.tides)
        ps[0].params["tctl_k1"] = 0.023 # in AU
        ps[0].params["tctl_tau"] = 0.3
        ps[0].params["Omega"] = 0.

        self.q = (ps[1].m/ps[0].m)
        self.T = ps[0].r**3/self.sim.G/ps[0].m/ps[0].params["tctl_tau"]
        self.taua = self.T/6/ps[0].params["tctl_k1"]/self.q/(1+self.q)*(ps[1].a/ps[0].r)**8

    def test_adamping(self):
        ps = self.sim.particles
        tmax = 4e4*ps[1].P
        apred = ps[0].r*((ps[1].a/ps[0].r)**8 - 48.*ps[0].params["tctl_k1"]*self.q*(1+self.q)*tmax/self.T)**(1./8.)

        self.sim.integrate(tmax)
        self.assertLess(abs((ps[1].a-apred)/apred), 1.e-2) # 1%
    
    def test_linear_adamping(self):
        ps = self.sim.particles
        tmax = self.taua/100
        apred = ps[1].a*np.exp(-tmax/self.taua)

        self.sim.integrate(tmax)
        self.assertLess(abs((ps[1].a-apred)/apred), 0.1) # 10%
    
    def test_linear_edamping(self):
        ps = self.sim.particles
        tmax = self.taua/100
        epred = ps[1].e*np.exp(-tmax/(6./27.*self.taua))

        self.sim.integrate(tmax)
        self.assertLess(abs((ps[1].e-epred)/epred), 0.1) # 10%
    
    def test_adamping_movecom(self):
        self.sim.particles[0].vy += 0.01 
        self.sim.particles[1].vy += 0.01 
        self.test_adamping()
    
    def test_linear_adamping_movecom(self):
        self.sim.particles[0].vy += 0.01 
        self.sim.particles[1].vy += 0.01 
        self.test_linear_adamping()
    
    def test_linear_edamping_movecom(self):
        self.sim.particles[0].vy += 0.01 
        self.sim.particles[1].vy += 0.01 
        self.test_linear_edamping()

if __name__ == '__main__':
    unittest.main()



import rebound
import reboundx
import unittest

def make():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(a=0.1, e=0.1)
    rebx = reboundx.Extras(sim)
    return sim, rebx
        
def makegr():
    sim, rebx = make()
    gr = rebx.load_force('gr')
    rebx.add_force(gr)
    gr.params['c'] = 1e2
    return sim, rebx

def dropsim():
    sim, rebx = makegr()
    return rebx

def droprebx():
    sim, rebx = makegr()
    return sim

def testEffect(reb_sim):
    reb_sim.contents.particles[0].params['c'] = 1.
    
class TestSegFaults(unittest.TestCase):
    def test_detach_preserves_force(self):
        sim, rebx = make()
        sim.additional_forces = testEffect
        addmass = rebx.load_operator('modify_mass')
        rebx.add_operator(addmass, dtfraction=1.,  timing="pre")
        rebx.detach(sim)
        sim.step()
        self.assertEqual(sim.particles[0].params['c'], 1)
    
    def test_detach_preserves_pretm(self):
        sim, rebx = make()
        sim.pre_timestep_modifications = testEffect
        addmass = rebx.load_operator('modify_mass')
        rebx.add_operator(addmass, dtfraction=1.,  timing="post")
        rebx.detach(sim)
        sim.step()
        self.assertEqual(sim.particles[0].params['c'], 1)
    
    def test_detach_preserves_pretm(self):
        sim, rebx = make()
        sim.post_timestep_modifications = testEffect
        gr = rebx.load_force('gr')
        rebx.add_force(gr)
        rebx.detach(sim)
        sim.step()
        self.assertEqual(sim.particles[0].params['c'], 1)

    def test_detach(self):
        sim, rebx = makegr()
        rebx.detach(sim)
        sim.integrate(10)
        self.assertLess(sim.particles[1].pomega, 1.e-10) # should be 0

    def test_create_in_function(self):
        sim, rebx = makegr()
        sim.integrate(10)
        self.assertGreater(sim.particles[1].pomega, 0.01)

    def test_droprebx(self):
        sim = droprebx()
        sim.integrate(10)
        self.assertGreater(sim.particles[1].pomega, 0.01)
    
    def test_dropsim(self):
        rebx = dropsim()
        with self.assertRaises(AttributeError):
            rebx.load_force("gr")

    def test_delete(self):
        for i in range(10):
            sim, rebx = makegr()
            sim.integrate(10)
        self.assertGreater(sim.particles[1].pomega, 0.01)

    def test_rebx_not_attached(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        with self.assertRaises(AttributeError):
            sim.particles[0].params["a"] = 7

if __name__ == '__main__':
    unittest.main()



"""
These are structures anad functions mainly used for testing from the Python side and for use in unit tests.
"""

from . import clibreboundx
from ctypes import c_void_p, c_int, c_long, Structure, byref, c_char_p
import warnings
from reboundx.extras import REBX_BINARY_WARNINGS

REBX_BINARY_FIELD_TYPE = {
        0: 'None',
        1: 'Operator',
        2: 'Particle',
        3: 'Rebx Structure',
        4: 'Param',
        5: 'Name',
        6: 'Param type',
        7: 'Param valaue',
        8: 'END',
        9: 'Particle index',
        10: 'Rebx integraator',
        11: 'Force type',
        12: 'Operator type',
        13: 'Step',
        14: 'Step dt fraction',
        15: 'Registered Param',
        16: 'Additional force',
        17: 'Param list',
        18: 'Registered Params',
        19: 'Allocated forces',
        20: 'Allocated operators',
        21: 'Additional forces',
        22: 'Pre timestep modifications',
        23: 'Post timestep modifications',
        24: 'Particles',
        25: 'Force',
        26: 'Snapshot',
        }

class BinaryField(Structure):
    _fields_ =  [  ("_type", c_int),
                    ("size", c_long)]
    @property
    def type(self):
        return REBX_BINARY_FIELD_TYPE[self._type]

    def __repr__(self):
        return 'Type: {0}, Size: {1}'.format(self.type, self.size)

def inspect_binary(filename):
    w = c_int(0)
    clibreboundx.rebx_input_inspect_binary.restype = c_void_p
    inf = clibreboundx.rebx_input_inspect_binary(c_char_p(filename.encode('ascii')), byref(w))
    for majorerror, value, message in REBX_BINARY_WARNINGS:
        if w.value & value:
            if majorerror:
                raise RuntimeError(message)
            else:
                warnings.warn(message, RuntimeWarning)
    return inf

def read_binary_field(inf):
    clibreboundx.rebx_input_read_binary_field.restype = BinaryField
    return clibreboundx.rebx_input_read_binary_field(c_void_p(inf))

def skip_binary_field(inf, size):
    clibreboundx.rebx_input_skip_binary_field(c_void_p(inf), size)




import rebound
from collections import MutableMapping
from .extras import Param, Node, Force, Operator, Extras, REBX_CTYPES
from . import clibreboundx
from ctypes import byref, c_double, c_int, c_int32, c_int64, c_uint, c_uint32, c_longlong, c_char_p, POINTER, cast
from ctypes import c_void_p, memmove, sizeof, addressof
import numpy as np
from rebound.tools import hash as rebhash

class Params(MutableMapping):
    def __init__(self, parent):
        self.verbose = 0        # set to 1 to diagnose problems
        self.parent = parent    # Particle, Force, Operator. Will work with any ctypes.Structure with appropriate ._sim and .ap fields

        offset = type(parent).ap.offset # Need this hack to get address of initially NULL ap ptr. See my stackoverflow
        self.ap = (c_void_p).from_buffer(parent, offset)
      
        # We want to be able to access params from objects like particles, objects etc. These must somehow point back to rebx for memory management
        # We can't add a rebx pointer in rebound.Particle, but can use its sim pointer. So we add a sim pointer to all rebx objects that need params.
        extrasvp = parent._sim.contents.extras
        if not extrasvp: # .extras = None
            raise AttributeError("Need to attach reboundx.Extras instance to simulation before setting params.")
        else:
            self.rebx = cast(extrasvp, POINTER(Extras))

    def __getitem__(self, key):
        param_type = clibreboundx.rebx_get_type(self.rebx, c_char_p(key.encode('ascii')))
        ctype = REBX_CTYPES[param_type]
        if ctype == None:
            raise AttributeError("REBOUNDx Error: Parameter '{0}' not found in REBOUNDx. Need to register it first.".format(key))
        
        clibreboundx.rebx_get_param.restype = c_void_p
        valptr = clibreboundx.rebx_get_param(self.rebx, self.ap, c_char_p(key.encode('ascii')))

        if ctype == c_void_p: # Don't know how to cast it, so return for user to cast
            if valptr is None:
                raise AttributeError("REBOUNDx Error: Parameter '{0}' not found on object.".format(key))
            return valptr
        
        valptr = cast(valptr, POINTER(ctype))
        try:
            val = valptr.contents.value # return python int or float rather than c_int or c_double
        except AttributeError:
            val = valptr.contents # Structure, return ctypes object
        except ValueError: # NULL access
            raise AttributeError("REBOUNDx Error: Parameter '{0}' not found on object.".format(key))

        return val

    def __setitem__(self, key, value):
        param_type = clibreboundx.rebx_get_type(self.rebx, c_char_p(key.encode('ascii')))
        ctype = REBX_CTYPES[param_type]
        if ctype == None:
            raise AttributeError("REBOUNDx Error: Parameter '{0}' not found in REBOUNDx. Need to register it first.".format(key))
        if ctype == c_double:
            clibreboundx.rebx_set_param_double(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), c_double(value))
        if ctype == c_int:
            clibreboundx.rebx_set_param_int(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), c_int(value))
        if ctype == c_uint32:
            clibreboundx.rebx_set_param_uint32(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), value)

        if ctype == Force:
            if not isinstance(value, Force):
                raise AttributeError("REBOUNDx Error: Parameter '{0}' must be assigned a Force object.".format(key))
            clibreboundx.rebx_set_param_pointer(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), byref(value))

        if ctype == rebound.Orbit:
            if not isinstance(value, rebound.Orbit):
                raise AttributeError("REBOUNDx Error: Parameter '{0}' must be assigned an Orbit object.".format(key))
            clibreboundx.rebx_set_param_pointer(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), byref(value))
        if ctype == c_void_p:
            clibreboundx.rebx_set_param_pointer(self.rebx, byref(self.ap), c_char_p(key.encode('ascii')), byref(value))

    def __delitem__(self, key):
        raise AttributeError("REBOUNDx Error: Removing particle params not implemented.")

    def __iter__(self):
        raise AttributeError("REBOUNDx Error: Iterator for params not implemented.")

    def __len__(self):
        clibreboundx.rebx_len.restype = c_int
        return clibreboundx.rebx_len(self.ap)



import rebound
import reboundx

class SimulationArchive(rebound.SimulationArchive):
    """
    SimulationArchive Class.
    """
    def __init__(self, filename, rebxfilename, *args, **kwargs):
        """
        Arguments
        ---------
        filename : str
            Filename of the SimulationArchive file to be opened.
        rebxfilename : str
            Filename of the REBOUNDx binary file.
        """
        super(SimulationArchive, self).__init__(filename, *args, **kwargs)
        self.rebxfilename = rebxfilename
        sim, rebx = self[0] # test you can open rebxfilename to warn user if not

    def __getitem__(self, key):
        sim = super(SimulationArchive, self).__getitem__(key)
        rebx = reboundx.Extras(sim, self.rebxfilename)
        return sim, rebx

    def getSimulation(self, *args, **kwargs):
        sim = super(SimulationArchive, self).getSimulation(*args, **kwargs)
        rebx = reboundx.Extras(sim, self.rebxfilename)
        return sim, rebx



import rebound
import reboundx
import unittest
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class TestMachineIndependent(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1., e=0.2)
        self.rebx = reboundx.Extras(self.sim)

    def test_modify_orbits_forces(self):
        name = "modify_orbits_forces"
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            try:
                sim = rebound.Simulation('binaries/twoplanets.bin')
            except:
                sim = rebound.Simulation()
                sim.add(m=1.)
                sim.add(m=1.e-4, a=1., e=0.1)
                sim.add(m=1.e-4, a=2, e=0.1, inc=0.2)
                sim.integrator="whfast"
                sim.dt = sim.particles[1].P/100
                sim.move_to_com()
                sim.save('binaries/twoplanets.bin')
           
            sim.automateSimulationArchive('binaries/modify_orbits_forces.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            mod = rebx.load_force('binaries/modify_orbits_forces')
            rebx.add_force(mod)
            ps = sim.particles
            ps[1].params['tau_a'] = -1e4
            ps[1].params['tau_e'] = -1e3
            ps[2].params['tau_e'] = -1e3
            ps[2].params['tau_inc'] = -1e3
            rebx.save('binaries/modify_orbits_forces.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/modify_orbits_forces.sa', 'binaries/modify_orbits_forces.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)

    def test_gr_full(self):
        name = 'gr_full'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            force.params['c'] = 1.e4
            ps = sim.particles
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)
    
    def test_gr(self):
        name = 'gr'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            force.params['c'] = 1.e4
            ps = sim.particles
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)
    
    def test_gr_potential(self):
        name = 'gr_potential'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            force.params['c'] = 1.e4
            ps = sim.particles
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)

    def test_radiation_forces(self):
        name = 'radiation_forces'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            force.params['c'] = 1.e4
            ps = sim.particles
            ps[1].params['beta'] = 0.5
            ps[2].params['beta'] = 0.3
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)

    def test_central_force(self):
        name = 'central_force'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            ps = sim.particles
            ps[0].params['Acentral'] = 1.e-3
            ps[0].params['gammacentral'] = -1
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)
    
    def test_gravitational_harmonics(self):
        name = 'gravitational_harmonics'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            force = rebx.load_force(name)
            rebx.add_force(force)
            ps = sim.particles
            ps[0].params['J2'] = 1.e-3
            ps[0].params['J4'] = 1.e-3
            ps[0].params['R_eq'] = 1.e-3
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)

    def test_modify_mass(self):
        name = 'modify_mass'
        try:
            rebbin = os.path.join(THIS_DIR, 'binaries/'+name+'.sa')
            rebxbin = os.path.join(THIS_DIR, 'binaries/'+name+'.rebx')
            sa = reboundx.SimulationArchive(rebbin, rebxbin)
        except:
            sim = rebound.Simulation('binaries/twoplanets.bin')
            sim.automateSimulationArchive('binaries/'+name+'.sa', interval=1e3, deletefile=True)
            rebx = reboundx.Extras(sim)
            mod = rebx.load_operator(name)
            rebx.add_operator(mod)
            ps = sim.particles
            ps[0].params['tau_mass'] = -1e4
            rebx.save('binaries/'+name+'.rebx')
            sim.integrate(1.e4)
            sa = reboundx.SimulationArchive('binaries/'+name+'.sa', 'binaries/'+name+'.rebx')

        simf, rebx = sa[-1]
        sim,  rebx = sa[0]
        sim.integrate(simf.t)
        self.assertEqual(sim.particles[0].x, simf.particles[0].x)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# Call this after adding a new effect source file, e.g., effect.c.  It takes all the .c files in reboundx/src and adds them to all the Makefiles where needed

import glob
import os
import sys

sources = [each for each in os.listdir('../src/') if each.endswith('.c')] 
print("Adding {0}".format(sources))
sourcestring = "SOURCES="
sourcesExamples = "SOURCES="
sourcesSetup = "        sources = ["
for source in sources:
    sourcestring += source + " "
    sourcesExamples += "$(REBOUNDXDIR)" + source + " "
    sourcesSetup += ' \'src/' + source + '\','
    
sourcestring += "\n"
sourcesExamples += "\n"
sourcesSetup = sourcesSetup[:-1] + '],\n' # remove comma after last source

with open("../src/Makefile") as f:
    lines = f.readlines()
    for i,l in enumerate(lines):
        if "modify_orbits_direct.c" in l:
            lines[i] = sourcestring

    with open("../src/Makefile", "w") as f:
        f.writelines(lines)

with open("../examples/Makefile") as f:
    lines = f.readlines()
    for i,l in enumerate(lines):
        if "modify_orbits_direct.c" in l:
            lines[i] = sourcesExamples

    with open("../examples/Makefile", "w") as f:
        f.writelines(lines)
'''
for example in os.listdir('../examples/'):
    try:
        with open("../examples/" + example + "/Makefile") as f:
            lines = f.readlines()
            for i,l in enumerate(lines):
                if "modify_orbits_direct.c" in l:
                    print('here')
                    lines[i] = sourcesExamples

            with open("../examples/" + example + "/Makefile", "w") as f:
                f.writelines(lines)
    except:
        pass
'''
with open("../setup.py") as f:
    setuplines = f.readlines()
    for i,l in enumerate(setuplines):
        if 'src/' in l:
            setuplines[i] = sourcesSetup

    with open("../setup.py", "w") as f:
        f.writelines(setuplines)




import rebound
import reboundx
import unittest
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
binary = os.path.join(THIS_DIR, 'binaries/twoplanets.bin')

class TestConservation(unittest.TestCase):
    def test_gr_full(self):
        name = 'gr_full'
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        force = rebx.load_force(name)
        rebx.add_force(force)
        force.params['c'] = 1.e4
        ps = sim.particles
        H0 = rebx.gr_full_hamiltonian(force)
        sim.integrate(1.e4)
        H = rebx.gr_full_hamiltonian(force)
        self.assertLess(abs((H-H0)/H0), 1.e-12)
    
    def test_gr(self):
        name = 'gr'
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        force = rebx.load_force(name)
        rebx.add_force(force)
        force.params['c'] = 1.e4
        ps = sim.particles
        H0 = rebx.gr_hamiltonian(force)
        sim.integrate(1.e4)
        H = rebx.gr_hamiltonian(force)
        self.assertLess(abs((H-H0)/H0), 1.e-12)
    
    def test_gr_potential(self):
        name = 'gr_potential'
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        force = rebx.load_force(name)
        rebx.add_force(force)
        force.params['c'] = 1.e4
        ps = sim.particles
        H0 = sim.calculate_energy() + rebx.gr_potential_potential(force)
        sim.integrate(1.e4)
        H = sim.calculate_energy() + rebx.gr_potential_potential(force)
        self.assertLess(abs((H-H0)/H0), 1.e-12)

    def test_central_force(self):
        name = 'central_force'
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        force = rebx.load_force(name)
        rebx.add_force(force)
        ps = sim.particles
        ps[0].params['Acentral'] = 1.e-4
        ps[0].params['gammacentral'] = -1
        H0 = sim.calculate_energy() + rebx.central_force_potential()
        sim.integrate(1.e4)
        H = sim.calculate_energy() + rebx.central_force_potential()
        self.assertLess(abs((H-H0)/H0), 1.e-12)

    def test_gravitational_harmonics(self):
        name = 'gravitational_harmonics'
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        force = rebx.load_force(name)
        rebx.add_force(force)
        ps = sim.particles
        ps[0].params['J2'] = 1.e-3
        ps[0].params['J4'] = 1.e-3
        ps[0].params['R_eq'] = 1.e-3
        H0 = sim.calculate_energy() + rebx.gravitational_harmonics_potential()
        sim.integrate(1.e4)
        H = sim.calculate_energy() + rebx.gravitational_harmonics_potential()
        self.assertLess(abs((H-H0)/H0), 1.e-12)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# This script automatically creates a list of examples by reading the header in all problem.c files.
import glob
import subprocess
ghash = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()

with open("version.txt") as f:
    reboundxversion = f.readlines()[0].strip()
    print("Updating version to "+reboundxversion)

with open("doc/index.rst") as f:
    index = f.readlines()

with open("doc/index.rst","w") as f:
    for i in range(0,len(index)):
        if "Welcome to REBOUNDx" in index[i]:
            index[i] = "Welcome to REBOUNDx ("+reboundxversion+")\n"
            underline = ""
            for j in range(len(index[i])-1):
                underline += "="
            underline += "\n"
            index[i+1] = underline
        f.write(index[i])

with open("README.rst") as f:
    readme = f.readlines()

with open("README.rst","w") as f:
    for i in range(0,len(readme)):
        if "badge/REBOUNDx-v" in readme[i]:
            readme[i] = ".. image:: https://img.shields.io/badge/REBOUNDx-v"+reboundxversion+"-green.svg?style=flat\n"
        f.write(readme[i])

with open("src/core.c") as f:
    reboundxlines = f.readlines()
    for i,l in enumerate(reboundxlines):
        if "**VERSIONLINE**" in l:
            reboundxlines[i] = "const char* rebx_version_str = \""+reboundxversion+"\";         // **VERSIONLINE** This line gets updated automatically. Do not edit manually.\n"

    with open("src/core.c", "w") as f:
        f.writelines(reboundxlines)

with open("setup.py") as f:
    setuplines = f.readlines()
    for i,l in enumerate(setuplines):
        if "version='" in l:
            setuplines[i] = "    version='"+reboundxversion+"',\n"
        if "GITHASHAUTOUPDATE" in l:
            setuplines[i] = "    ghash_arg = \"-DREBXGITHASH="+ghash+"\" #GITHASHAUTOUPDATE\n"

    with open("setup.py", "w") as f:
        f.writelines(setuplines)

shortversion = reboundxversion
while shortversion[-1] is not '.':
    shortversion = shortversion[:-1]
    
shortversion = shortversion[:-1]

with open("doc/conf.py") as f:
    conflines = f.readlines()
    for i,l  in enumerate(conflines):
        if "version =" in l:
            conflines[i] = "version = '"+shortversion+"'\n"
        if "release =" in l:
            conflines[i] = "release = '"+reboundxversion+"'\n"

    with open("doc/conf.py", "w") as f:
        f.writelines(conflines)



from . import clibreboundx
from ctypes import Structure, c_double, POINTER, c_int, c_uint, c_long, c_ulong, c_void_p, c_char_p, CFUNCTYPE, byref, c_uint32, c_uint, cast, c_char
import rebound
import reboundx
import warnings

integrators = {"implicit_midpoint": 0, "rk4":1, "euler": 2, "rk2": 3, "none": -1}

REBX_TIMING = {"pre":-1, "post":1}
REBX_FORCE_TYPE = {"none":0, "pos":1, "vel":2}
REBX_OPERATOR_TYPE = {"none":0, "updater":1, "recorder":2}

REBX_BINARY_WARNINGS = [
    (True, 1, "REBOUNDx: Cannot open binary file. Check filename."),
    (True, 2, "REBOUNDx: Binary file is unreadable. Please open an issue on Github mentioning the version of REBOUND and REBOUNDx you are using and include the binary file."),
    (True, 4, "REBOUNDx: Ran out of system memory."),
    (True, 8, "REBOUNDx: REBOUNDx structure couldn't be loaded."),
    (True, 16, "REBOUNDx: At least one registered parameter was not loaded. This typically indicates the binary is corrupt or was saved with an incompatible version to the current one being used."),
    (False, 32, "REBOUNDx: At least one force or operator parameter was not loaded from the binary file. This typically indicates the binary is corrupt or was saved with an incompatible version to the current one being used."),
    (False, 64, "REBOUNDx: At least one particle's parameters were not loaded from the binary file."),
    (False, 128, "REBOUNDx: At least one force was not loaded from the binary file. If binary was created with a newer version of REBOUNDx, a particular force may not be implemented in your current version of REBOUNDx."),
    (False, 256, "REBOUNDx: At least one operator was not loaded from the binary file. If binary was created with a newer version of REBOUNDx, a particular force may not be implemented in your current version of REBOUNDx."),
    (False, 512, "REBOUNDx: At least one operator step was not loaded from the binary file."),
    (False, 1024, "REBOUNDx: At least one force was not added to the simulation. If binary was created with a newer version of REBOUNDx, a particular force may not be implemented in your current version of REBOUNDx."),
    (False, 2048, "REBOUNDx: Unknown field found in binary file. Any unknown fields not loaded.  This can happen if the binary was created with a later version of REBOUNDx than the one used to read it."),
    (False, 4096, "REBOUNDx: Unknown list in the REBOUNDx structure wasn't loaded. This can happen if the binary was created with a later version of REBOUNDx than the one used to read it."),
    (False, 8192, "REBOUNDx: The value of at least one parameter was not loaded. This can happen if a custom structure was added by the user as a parameter. See Parameters.ipynb jupyter notebook example."),
    (False,16384, "REBOUNDx: Binary file was saved with a different version of REBOUNDx. Binary format might have changed. Check that effects and parameters are loaded as expected.")
]

class Extras(Structure):
    """
    Main object used for all REBOUNDx operations, tied to a particular REBOUND simulation.
    This is an abstraction of the C struct rebx_extras, with all the C convenience functions
    and functions for adding effects implemented as methods of the class.  
    The fastest way to understand it is to follow the examples at :ref:`ipython_examples`.  
    """
    
    def __new__(cls, sim, filename=None):
        rebx = super(Extras,cls).__new__(cls)
        return rebx

    def __init__(self, sim, filename=None):
        sim._extras_ref = self # add a reference to this instance in sim to make sure it's not garbage collected_ 
        clibreboundx.rebx_initialize(byref(sim), byref(self))
        # Create simulation
        if filename==None:
            # Create a new rebx instance
           clibreboundx.rebx_register_default_params(byref(self))
        else:
            # Recreate existing simulation.
            # Load registered parameters from binary
            w = c_int(0)
            clibreboundx.rebx_init_extras_from_binary(byref(self), c_char_p(filename.encode('ascii')), byref(w))
            for majorerror, value, message in REBX_BINARY_WARNINGS:
                if w.value & value:
                    if majorerror:
                        raise RuntimeError(message)
                    else:
                        warnings.warn(message, RuntimeWarning)
        self.process_messages()

    def __del__(self):
        if self._b_needsfree_ == 1:
            clibreboundx.rebx_free_pointers(byref(self))

    def detach(self, sim):
        sim._extras_ref = None # remove reference to rebx so it can be garbage collected 
        clibreboundx.rebx_detach(byref(sim), byref(self))
    
    #######################################
    # Functions for manipulating REBOUNDx effects
    #######################################

    def register_param(self, name, param_type):
        type_enum = REBX_C_PARAM_TYPES[param_type] 
        clibreboundx.rebx_register_param(byref(self), c_char_p(name.encode('ascii')), c_int(type_enum))
        self.process_messages()

    def load_force(self, name):
        clibreboundx.rebx_load_force.restype = POINTER(Force)
        ptr = clibreboundx.rebx_load_force(byref(self), c_char_p(name.encode('ascii')))
        self.process_messages()
        return ptr.contents
    
    def create_force(self, name):
        clibreboundx.rebx_create_force.restype = POINTER(Force)
        ptr = clibreboundx.rebx_create_force(byref(self), c_char_p(name.encode('ascii')))
        self.process_messages()
        return ptr.contents

    def load_operator(self, name):
        clibreboundx.rebx_load_operator.restype = POINTER(Operator)
        ptr = clibreboundx.rebx_load_operator(byref(self), c_char_p(name.encode('ascii')))
        self.process_messages()
        return ptr.contents

    def create_operator(self, name):
        clibreboundx.rebx_create_operator.restype = POINTER(Operator)
        ptr = clibreboundx.rebx_create_operator(byref(self), c_char_p(name.encode('ascii')))
        self.process_messages()
        return ptr.contents

    def add_force(self, force):
        if not isinstance(force, reboundx.extras.Force):
            raise TypeError("REBOUNDx Error: Object passed to rebx.add_force is not a reboundx.Force instance.")
        clibreboundx.rebx_add_force(byref(self), byref(force))
        self.process_messages()

    def add_operator(self, operator, dtfraction=None, timing="post"):
        if not isinstance(operator, reboundx.extras.Operator):
            raise TypeError("REBOUNDx Error: Object passed to rebx.add_operator is not a reboundx.Operator instance.")
        if dtfraction is None:
            clibreboundx.rebx_add_operator(byref(self), byref(operator))
        else:
            timingint = REBX_TIMING[timing]
            clibreboundx.rebx_add_operator_step(byref(self), byref(operator), c_double(dtfraction), c_int(timingint))
        self.process_messages()

    def get_force(self, name):
        clibreboundx.rebx_get_force.restype = POINTER(Force)
        ptr = clibreboundx.rebx_get_force(byref(self), c_char_p(name.encode('ascii')))
        if ptr:
            return ptr.contents
        else:
            raise AttributeError("REBOUNDx Error: Force {0} passed to rebx.get_force not found.".format(name))

    def get_operator(self, name):
        clibreboundx.rebx_get_operator.restype = POINTER(Operator)
        ptr = clibreboundx.rebx_get_operator(byref(self), c_char_p(name.encode('ascii')))
        self.process_messages()
        if ptr:
            return ptr.contents
        else:
            raise AttributeError("REBOUNDx Error: Operator {0} passed to rebx.get_operator not found.".format(name))

    def remove_force(self, force):
        if not isinstance(force, reboundx.extras.Force):
            raise TypeError("REBOUNDx Error: Object passed to rebx.remove_force is not a reboundx.Force instance.")
        success = clibreboundx.rebx_remove_force(byref(self), byref(force))
        if not success:
            raise AttributeError("REBOUNDx Error: Force {0} passed to rebx.remove_force not found in simulation.")

    def remove_operator(self, operator):
        if not isinstance(operator, reboundx.extras.Operator):
            raise TypeError("REBOUNDx Error: Object passed to rebx.remove_operator is not a reboundx.Operator instance.")
        success = clibreboundx.rebx_remove_operator(byref(self), byref(operator))
        if not success:
            raise AttributeError("REBOUNDx Error: Operator {0} passed to rebx.remove_operator not found in simulation.")

    #######################################
    # Input/Output Routines
    #######################################
    def save(self, filename):
        """
        Save the entire REBOUND simulation to a binary file.
        """
        clibreboundx.rebx_output_binary(byref(self), c_char_p(filename.encode("ascii")))
        self.process_messages()

    #######################################
    # Convenience Functions
    #######################################

    def rad_calc_beta(self, G, c, source_mass, source_luminosity, radius, density, Q_pr):
        clibreboundx.rebx_rad_calc_beta.restype = c_double
        return clibreboundx.rebx_rad_calc_beta(c_double(G), c_double(c), c_double(source_mass), c_double(source_luminosity), c_double(radius), c_double(density), c_double(Q_pr))

    def rad_calc_particle_radius(self, G, c, source_mass, source_luminosity, beta, density, Q_pr):
        clibreboundx.rebx_rad_calc_particle_radius.restype = c_double
        return clibreboundx.rebx_rad_calc_particle_radius(c_double(G), c_double(c), c_double(source_mass), c_double(source_luminosity), c_double(beta), c_double(density), c_double(Q_pr))

    def central_force_Acentral(self, p, primary, pomegadot, gamma):
        clibreboundx.rebx_central_force_Acentral.restype = c_double
        Acentral = clibreboundx.rebx_central_force_Acentral(p, primary, c_double(pomegadot), c_double(gamma))
        self.process_messages()
        return Acentral

    # Hamiltonian calculation functions
    def gr_full_hamiltonian(self, force):
        clibreboundx.rebx_gr_full_hamiltonian.restype = c_double
        return clibreboundx.rebx_gr_full_hamiltonian(byref(self), byref(force))
    
    def gr_hamiltonian(self, force):
        clibreboundx.rebx_gr_hamiltonian.restype = c_double
        return clibreboundx.rebx_gr_hamiltonian(byref(self), byref(force))
    
    # Potential calculation functions
    def gr_potential_potential(self, force):
        clibreboundx.rebx_gr_potential_potential.restype = c_double
        return clibreboundx.rebx_gr_potential_potential(byref(self), byref(force))
    
    def tides_constant_time_lag_potential(self, force):
        clibreboundx.rebx_tides_constant_time_lag_potential.restype = c_double
        return clibreboundx.rebx_tides_constant_time_lag_potential(byref(self), byref(force))

    def central_force_potential(self):
        clibreboundx.rebx_central_force_potential.restype = c_double
        return clibreboundx.rebx_central_force_potential(byref(self))
    
    def gravitational_harmonics_potential(self):
        clibreboundx.rebx_gravitational_harmonics_potential.restype = c_double
        return clibreboundx.rebx_gravitational_harmonics_potential(byref(self))

    def process_messages(self):
        try:
            self._sim.contents.process_messages()
        except ValueError: # _sim is NULL
            raise AttributeError("REBOUNDx Error: The Simulation instance REBOUNDx was attached to no longer exists. This can happen if the Simulation instance goes out of scope or otherwise gets garbage collected.")
#################################################
# Generic REBOUNDx definitions
#################################################


class Param(Structure): # need to define fields afterward because of circular ref in linked list
    pass    
Param._fields_ =  [ ("name", c_char_p),
                    ("type", c_int),
                    ("value", c_void_p)]

class Node(Structure): # need to define fields afterward because of circular ref in linked list
    pass    
Node._fields_ =  [  ("object", c_void_p),
                    ("next", POINTER(Node))]

class Operator(Structure):
    @property
    def operator_type(self):
        return self._operator_type

    @operator_type.setter
    def operator_type(self, value):
        self._operator_type = REBX_OPERATOR_TYPE[value.lower()]

    @property
    def step_function(self):
        return self._step_function

    @step_function.setter
    def step_function(self, func):
        self._sfp = STEPFUNCPTR(func) # keep a reference to func so it doesn't get garbage collected
        self._step_function = self._sfp

    def step(self, sim, dt):
        self._step_function(byref(sim), byref(self), dt)

    @property 
    def params(self):
        params = Params(self)
        return params

STEPFUNCPTR = CFUNCTYPE(None, POINTER(rebound.Simulation), POINTER(Operator), c_double)

Operator._fields_ = [   ("name", c_char_p),
                        ("ap", POINTER(Node)),
                        ("_sim", POINTER(rebound.Simulation)),
                        ("_operator_type", c_int),
                        ("_step_function", STEPFUNCPTR)]
class Force(Structure):
    @property
    def force_type(self):
        return self._force_type

    @force_type.setter
    def force_type(self, value):
        self._force_type = REBX_FORCE_TYPE[value.lower()]

    @property
    def update_accelerations(self):
        return self._update_accelerations

    @update_accelerations.setter
    def update_accelerations(self, func):
        self._ffp = FORCEFUNCPTR(func) # keep a reference to func so it doesn't get garbage collected
        self._update_accelerations = self._ffp

    @property 
    def params(self):
        params = Params(self)
        return params

FORCEFUNCPTR = CFUNCTYPE(None, POINTER(rebound.Simulation), POINTER(Force), POINTER(rebound.Particle), c_int)

Force._fields_ = [  ("name", c_char_p),
                    ("ap", POINTER(Node)),
                    ("_sim", POINTER(rebound.Simulation)),
                    ("_force_type", c_int),
                    ("_update_accelerations", FORCEFUNCPTR)]

# Need to put fields after class definition because of self-referencing
Extras._fields_ =  [("_sim", POINTER(rebound.Simulation)),
                    ("_additional_forces", POINTER(Node)),
                    ("_pre_timestep_modifications", POINTER(Node)),
                    ("_post_timestep_modifications", POINTER(Node)),
                    ("_registered_params", POINTER(Node)),
                    ("_allocated_forces", POINTER(Node)),
                    ("_allocated_operators", POINTER(Node))]

class Interpolator(Structure):
    def __new__(cls, rebx, times, values, interpolation):
        interp = super(Interpolator, cls).__new__(cls)
        return interp 

    def __init__(self, rebx, times, values, interpolation="spline"):
        try:
            Nvalues = len(times)
            Nvalues2 = len(values)
        except:
            raise TypeError("REBOUNDx Error: Times and values passed to Interpolator must be lists or arrays")
        if Nvalues != Nvalues2:
            raise ValueError("REBOUNDx Error: Times and values must be same length)")

        interpolation = interpolation.lower()
        if interpolation in INTERPOLATION_TYPE:
            interp = INTERPOLATION_TYPE[interpolation]
        else:
            raise ValueError("REBOUNDx Error: Interpolation type not supported")

        DblArr = c_double * Nvalues
        clibreboundx.rebx_init_interpolator(byref(rebx), byref(self), c_int(Nvalues), DblArr(*times), DblArr(*values), c_int(interp)) 

    def interpolate(self, rebx, t):
        clibreboundx.rebx_interpolate.restype = c_double
        return clibreboundx.rebx_interpolate(byref(rebx), byref(self), c_double(t))
    
    def __del__(self):
        if self._b_needsfree_ == 1:
            clibreboundx.rebx_free_interpolator_pointers(byref(self))
        
Interpolator._fields_ = [  ("interpolation", c_int),
                    ("times", POINTER(c_double)),
                    ("values", POINTER(c_double)),
                    ("Nvalues", c_int),
                    ("y2", POINTER(c_double)),
                    ("klo", c_int)]

INTERPOLATION_TYPE = {"none":0, "spline":1}

# This list keeps pairing from C rebx_param_type enum to ctypes type 1-to-1. Derive the required mappings from it
REBX_C_TO_CTYPES = [["REBX_TYPE_NONE", None], ["REBX_TYPE_DOUBLE", c_double], ["REBX_TYPE_INT",c_int], ["REBX_TYPE_POINTER", c_void_p], ["REBX_TYPE_FORCE", Force], ["REBX_TYPE_UNIT32", c_uint32], ["REBX_TYPE_ORBIT", rebound.Orbit]]
REBX_CTYPES = {} # maps int value of rebx_param_type enum to ctypes type
REBX_C_PARAM_TYPES = {} # maps string of rebx_param_type enum to int
for i, pair in enumerate(REBX_C_TO_CTYPES):
    REBX_CTYPES[i] = pair[1]
    REBX_C_PARAM_TYPES[pair[0]] = i

from .params import Params



import rebound
import reboundx
from reboundx import data
import unittest
import math
import numpy as np
from ctypes import c_uint, c_uint8, c_uint32, c_uint64

def mycomp(obj1, obj2):
    if type(obj1) != type(obj2):
        return False
    for attr in [attr for attr in dir(obj1) if not attr.startswith('_')]:
        if getattr(obj1, attr) != getattr(obj2, attr):
            return False
    return True
        
from ctypes import Structure, c_double, cast, POINTER
class Mystruct(Structure):
    _fields_ = [('dt', c_double),
                ('c', c_double)]

class TestParams(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1.)
        self.rebx = reboundx.Extras(self.sim)
        self.gr = self.rebx.load_force("gr")
        self.mm = self.rebx.load_operator("modify_mass")
        self.p = self.sim.particles[1]

    def tearDown(self):
        self.sim = None

    def test_adddouble(self):
        self.gr.params['c'] = 1.2
        self.mm.params['c'] = 1.4
        self.p.params['c'] = 1.7
        self.assertAlmostEqual(self.gr.params["c"], 1.2, delta=1.e-15)
        self.assertAlmostEqual(self.mm.params["c"], 1.4, delta=1.e-15)
        self.assertAlmostEqual(self.p.params["c"], 1.7, delta=1.e-15)

    def test_updatedouble(self):
        self.gr.params['c'] = 1.2
        self.mm.params['c'] = 1.4
        self.p.params['c'] = 1.7
        
        self.gr.params['c'] = 2.2
        self.mm.params['c'] = 2.4
        self.p.params['c'] = 2.7
        self.assertAlmostEqual(self.gr.params["c"], 2.2, delta=1.e-15)
        self.assertAlmostEqual(self.mm.params["c"], 2.4, delta=1.e-15)
        self.assertAlmostEqual(self.p.params["c"], 2.7, delta=1.e-15)
    
    def test_updatedoublecopy(self):
        a = 1.2
        self.gr.params['c'] = a

        a = 3.7
        self.assertAlmostEqual(self.gr.params["c"], 1.2, delta=1.e-15) # shouldn't reflect update because copied
    
    def test_addint(self):
        self.gr.params['gr_source'] = 1
        self.mm.params['gr_source'] = 2
        self.p.params['gr_source'] = 3
        self.assertAlmostEqual(self.gr.params["gr_source"], 1, delta=1.e-15)
        self.assertAlmostEqual(self.mm.params["gr_source"], 2, delta=1.e-15)
        self.assertAlmostEqual(self.p.params["gr_source"], 3, delta=1.e-15)
    
    def test_updateint(self):
        self.gr.params['gr_source'] = 1
        self.mm.params['gr_source'] = 2 
        self.p.params['gr_source'] = 3 
        
        self.gr.params['gr_source'] = 11 
        self.mm.params['gr_source'] = 22
        self.p.params['gr_source'] = 33
        self.assertEqual(self.gr.params["gr_source"], 11)
        self.assertEqual(self.mm.params["gr_source"], 22)
        self.assertEqual(self.p.params["gr_source"], 33)
   
    def test_updateintcopy(self):
        a = 1
        self.gr.params['gr_source'] = a

        a = 3
        self.assertEqual(self.gr.params["gr_source"], 1) # shouldn't reflect update because copied
    
    def test_addforce(self):
        self.gr.params['c'] = 3.5
        self.p.params['force'] = self.gr
        newgr = self.p.params['force']
        self.assertAlmostEqual(newgr.params['c'], 3.5, delta=1.e-15)

    def test_addnonforce(self):
        with self.assertRaises(AttributeError):
            self.p.params['force'] = 3.

    def test_updateforceptr(self):
        self.gr.params['c'] = 3.5
        self.p.params['force'] = self.gr
        self.gr.params['c'] = 4.5
        newgr = self.p.params['force']
        self.assertAlmostEqual(newgr.params['c'], 4.5, delta=1.e-15)
    
    def test_updateforce(self):
        self.p.params['force'] = self.gr
        self.gr.params['c'] = 3.5

        newforce = self.rebx.load_force('gr')
        newforce.params['c'] = -3.5
        self.p.params['force'] = newforce
        newgr = self.p.params['force']
        self.assertAlmostEqual(newgr.params['c'], -3.5, delta=1.e-15)
    
    def test_addcustomstruct(self):
        self.rebx.register_param('my_new_struct', 'REBX_TYPE_POINTER')
        s = Mystruct()
        s.dt = 0.1
        s.c = 3.5
        self.gr.params['my_new_struct'] = s
        new_s = self.gr.params['my_new_struct']
        new_s = cast(new_s, POINTER(Mystruct)).contents
        self.assertAlmostEqual(new_s.dt, 0.1, delta=1.e-15)
        self.assertAlmostEqual(new_s.c, 3.5, delta=1.e-15)

    def test_updatecustomstructptr(self):
        self.rebx.register_param('my_new_struct', 'REBX_TYPE_POINTER')
        s = Mystruct()
        s.dt = 0.1
        s.c = 3.5
        self.gr.params['my_new_struct'] = s

        s.dt = 1.1
        new_s = self.gr.params['my_new_struct']
        new_s = cast(new_s, POINTER(Mystruct)).contents
        self.assertAlmostEqual(new_s.dt, 1.1, delta=1.e-15)
    
    def test_updatecustomstruct(self):
        self.rebx.register_param('my_new_struct', 'REBX_TYPE_POINTER')
        s = Mystruct()
        s.dt = 0.1
        s.c = 3.5
        self.gr.params['my_new_struct'] = s

        s2 = Mystruct()
        s2.dt = -0.1
        s2.c = -3.5
        self.gr.params['my_new_struct'] = s2

        new_s = self.gr.params['my_new_struct']
        new_s = cast(new_s, POINTER(Mystruct)).contents
        self.assertAlmostEqual(new_s.dt, -0.1, delta=1.e-15)
        self.assertAlmostEqual(new_s.c, -3.5, delta=1.e-15)

    def test_getnotregistered(self):
        with self.assertRaises(AttributeError):
            b = self.gr.params['asd;flkj']

    def test_notregistered(self):
        with self.assertRaises(AttributeError):
            self.gr.params['asldkjf'] = 1.2
    
    def test_registerexisting(self):
        with self.assertRaises(RuntimeError):
            self.rebx.register_param('c', 'REBX_TYPE_INT')

    def test_not_attached(self):
        with self.assertRaises(AttributeError):
            b = self.gr.params['beta']
    
    def test_custom_not_attached(self):
        self.rebx.register_param('my_custom', 'REBX_TYPE_POINTER')
        with self.assertRaises(AttributeError):
            b = self.gr.params['my_custom']

    def test_newdouble(self):
        self.rebx.register_param('my_new_double', 'REBX_TYPE_DOUBLE')
        self.gr.params['my_new_double'] = 1.2
        self.assertAlmostEqual(self.gr.params["my_new_double"], 1.2, delta=1.e-15)
    
    def test_newint(self):
        self.rebx.register_param('my_new_int', 'REBX_TYPE_INT')
        self.gr.params['my_new_int'] = 2
        self.assertEqual(self.gr.params["my_new_int"], 2)

    def test_length(self):
        self.gr.params['c'] = 1.3
        self.gr.params['gr_source'] = 7
        self.gr.params['tau_mass'] = 3.2
        self.assertEqual(len(self.gr.params), 3)

    def test_iter(self):
        with self.assertRaises(AttributeError):
            for p in self.gr.params:
                pass

    def test_del(self):
        with self.assertRaises(AttributeError):
            del self.gr.params["b"]

if __name__ == '__main__':
    unittest.main()



import rebound
import reboundx
import unittest
import os
import numpy as np

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
binary = os.path.join(THIS_DIR, 'binaries/twoplanets.bin')

class TestConservation(unittest.TestCase):
    def test_adiabatic_mass_loss(self):
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        times = [0, 2000., 4000., 6000., 8000., 10000.]
        values = [1., 0.8, 0.6, 0.4, 0.3, 0.2]
        starmass = reboundx.Interpolator(rebx, times, values, "spline")

        Nout = 1000
        ts = np.linspace(0., 1.e4, Nout)
        ps  = sim.particles
        a10 = ps[1].a
        for i, time in enumerate(ts):
            sim.integrate(time)
            ps[0].m = starmass.interpolate(rebx, t=sim.t)
            sim.move_to_com() # lost mass had momentum, so need to move back to COM frame
         
        self.assertLess(abs((ps[0].m-values[-1])/values[-1]), 1.e-6)
        self.assertLess(abs((ps[1].a-5*a10)/a10), 1.e-2)
   
    def test_klo(self):
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        times = [0, 2000., 4000., 6000., 8000., 10000.]
        values = [1., 0.8, 0.6, 0.4, 0.3, 0.2]
        starmass = reboundx.Interpolator(rebx, times, values, "spline")
        
        m0 = sim.particles[0].m
        mint0 = starmass.interpolate(rebx, t=0)

        self.assertLess(abs((m0-mint0)/m0), 1.e-6)

        mint1 = starmass.interpolate(rebx, t=5000) # advance klo
        mint1 = starmass.interpolate(rebx, t=0) # advance klo
        
        self.assertLess(abs((m0-mint1)/m0), 1.e-6)

    def test_backwards_time(self):
        sim = rebound.Simulation(binary)
        sim.integrator = "ias15"
        rebx = reboundx.Extras(sim)
        times = [0, 2000., 4000., 6000., 8000., 10000.]
        values = [1., 0.8, 0.6, 0.4, 0.3, 0.2]
        starmass = reboundx.Interpolator(rebx, times, values, "spline")

        Nout = 1000
        ts = np.linspace(0., 1.e4, Nout)
        ps  = sim.particles
        a10 = ps[1].a
        m0 = ps[0].m

        for i, time in enumerate(ts):
            sim.integrate(time)
            ps[0].m = starmass.interpolate(rebx, t=sim.t)
            sim.move_to_com() # lost mass had momentum, so need to move back to COM frame
        sim.dt = -0.001
        for i, time in enumerate(times[::-1]):
            sim.integrate(time)
            ps[0].m = starmass.interpolate(rebx, t=sim.t)
            sim.move_to_com() # lost mass had momentum, so need to move back to COM frame
        self.assertLess(abs((ps[0].m-m0)/m0), 1.e-2)
        self.assertLess(abs((ps[1].a-a10)/a10), 1.e-2)
    
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# REBOUNDx documentation build configuration file, created by
# sphinx-quickstart on Sun Nov 29 14:22:27 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import shlex
import subprocess
import glob
import re
from datetime import datetime

# Doxygen trigger
read_the_docs_build = os.environ.get('READTHEDOCS', None) == 'True'
if read_the_docs_build:
    subprocess.call('cd doxygen; doxygen', shell=True)

# C Example update
with open("c_examples.rst","w") as fd:
    fd.write(".. _c_examples:\n\n")
    fd.write("Examples (C)\n")
    fd.write("============\n\n")
    for problemc in glob.glob("../examples/*/problem.c"):
        will_output = 0
        with open(problemc) as pf:
            fd.write(".. _c_example_{0}:\n\n".format(problemc.split('/', 2)[-1].split('/',1)[0])) # make a label c_example_name_of_effect for cross-referencing
            did_output=0
            empty_lines = 0
            for line in pf:
                if line[0:3] == "/**":
                    will_output += 1
                if line[0:3] == " */":
                    will_output = -1
                    line = ""
                    fd.write("\n\n.. code-block:: c\n")
                if will_output>1:
                    if will_output == 2:
                        under = "-"*(len(line.strip())-2)
                        line = "  " +line.strip() + '\n' + under
                    will_output = 2
                    if len(line[3:].strip())==0:
                        fd.write("\n\n"+line[3:].strip())
                    else:
                        fd.write(line[3:].strip() + " " )
                if will_output==-1:
                    fd.write("   " +line.rstrip() + "\n" )
                    did_output = 1
                if will_output>0:
                    will_output += 1
            fd.write("\n\nThis example is located in the directory `examples/"+problemc.split("/")[2]+"`\n\n")
            if did_output==0:
                print("Warning: Did not find description in "+problemc)

# Update effects.rst
def cleanline(line):
    if line[3:] == "":
        return "\n"
    else:
        return line[3:]

sources = [each for each in os.listdir('../src') if each.endswith('.c')]
sources = [source[:-2] for source in sources] # all filenames ending with .c with the .c removed

with open("effects.rst", "w") as f:
    with open("effect_headers.txt", "r") as fh:
        lines = fh.readlines()
        i=0
        while i < len(lines):
            if lines[i].startswith(".. #"):
                i += 1    # skip comment line
            else:
                if "$$$" in lines[i]: # New category
                    i += 1
                    while lines[i].strip() == "": # skip white lines
                        i += 1
                    category = lines[i] # store category
                    while i < len(lines) and "$$$" not in lines[i]: # print category description
                        f.writelines(lines[i])
                        i += 1

                    for source in sources:
                        with open("../src/" + source + ".c", "r") as fs: # search sources that match category
                            sourcelines = fs.readlines()
                            j=0
                            while "#include" not in sourcelines[j]:
                                if "$$$" in sourcelines[j]: # start of documentation block
                                    j += 1
                                    while sourcelines[j].strip() == "*": # skip empty lines
                                        j += 1
                                    res = re.search(r'\$(.*)\$', cleanline(sourcelines[j])) # search for category between dollar signs
                                    if res is not None:
                                        cat = res.group(1)
                                        if cat.strip().lower() == category.strip().lower(): # source is in category, need to document
                                            j += 1
                                            f.write(".. _"+source+":\n\n") # write tag
                                            f.write(source+"\n") # write implementation heading
                                            for k in range(len(source)):
                                                f.write("*")
                                            f.write("\n")
                                            while " */" not in sourcelines[j]:
                                                f.write(cleanline(sourcelines[j]))
                                                j += 1
                                            f.write("\n")
                                        else: # category doesn't match
                                           break
                                    else: # first non-empty line doesn't have category
                                        break
                                else:
                                    j += 1

                else:
                    f.writelines(lines[i])
                    i += 1

        # Now document sources that had no category included separately
        for source in sources:
            with open("../src/" + source + ".c", "r") as fs: # search sources that match category
                sourcelines = fs.readlines()
                j=0
                while "#include" not in sourcelines[j]:
                    if "$$$" in sourcelines[j]: # start of documentation block
                        j += 1
                        while sourcelines[j].strip() == "*": # skip empty lines
                            j += 1
                        res = re.search(r'\$(.*)\$', cleanline(sourcelines[j])) # search for category between dollar signs
                        if res is None: # only document if no category present
                            f.write(source+"\n") # write section heading
                            for k in range(len(source)):
                                f.write("^")
                            f.write("\n\n")
                            f.write(".. _"+source+":\n\n") # write tag
                            f.write(source+"\n") # write implementation heading
                            for k in range(len(source)):
                                f.write("*")
                            f.write("\n\n")
                            while " */" not in sourcelines[j]:
                                f.write(cleanline(sourcelines[j]))
                                j += 1
                            break
                            f.write("\n")
                        else: # had category, don't document
                            break
                    else:
                        j += 1


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'breathe',
]
releases_github_path = "dtamayo/reboundx"
breathe_projects = { "libreboundx": "doxygen/xml/" }
breathe_default_project = "libreboundx"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '3.3'
# The full version, including alpha/beta/rc tags.
release = '3.3.1'

# General information about the project.
project = u"REBOUNDx ({0})".format(release)
year = datetime.now().year
author = 'Dan Tamayo, Hanno Rein'
copyright = u"{0} {1}".format(year, author)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "description": "A library for additional effects in REBOUND N-body integrations",
    "github_user": "dtamayo",
    "github_repo": "reboundx",
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}
html_sidebars = {
                '**': ['about.html', 'navigation.html', 'relations.html', 'searchbox.html']
                }

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'REBOUNDxdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'REBOUNDx.tex', 'REBOUNDx Documentation',
   'Dan Tamayo', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'reboundx', 'REBOUNDx Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'REBOUNDx', 'REBOUNDx Documentation',
   author, 'REBOUNDx', 'Library for additional physics in REBOUND simulations.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}



import rebound
import reboundx
import unittest

class TestRebx(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1., e=0.2)
        self.rebx = reboundx.Extras(self.sim)

    def test_simple(self):
        self.gr = self.rebx.load_force('gr')
        self.rebx.add_force(self.gr)
        self.gr.params['c'] = 1e2

        self.sim.integrate(10)
        self.assertGreater(self.sim.particles[1].pomega, 1.e-4)

if __name__ == '__main__':
    unittest.main()



import rebound
import reboundx
import unittest

class TestForces(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1., e=0.2)
        self.rebx = reboundx.Extras(self.sim)

    def test_loadforce(self):
        gr = self.rebx.load_force('gr')
    
    def test_loadforcewrongname(self):
        with self.assertRaises(RuntimeError):
            gr = self.rebx.load_force('gr2')

    def test_customforce(self):
        cust = self.rebx.create_force('myforce')
        def myforce(sim, force, particles, N):
            sim.contents.particles[1].ax += 1.e-4
        cust.update_accelerations = myforce
        cust.force_type = 'pos'
        self.rebx.add_force(cust)
        self.sim.integrate(10)
        self.assertGreater(self.sim.particles[1].pomega, -0.01)

    def test_customnoforce(self):
        cust = self.rebx.create_force('myforce')
        cust.force_type = 'pos'
        with self.assertRaises(RuntimeError):
            self.rebx.add_force(cust)

    def test_customnoforcetype(self):
        cust = self.rebx.create_force('myforce')
        def myforce(sim, force, particles, N):
            sim.contents.particles[1].ax += 1.e-4
        cust.update_accelerations = myforce
        with self.assertRaises(RuntimeError):
            self.rebx.add_force(cust)
    
    def test_getforce(self):
        gr = self.rebx.load_force('gr')
        gr.params['gr_source'] = 3
        newgr = self.rebx.get_force('gr')
        self.assertEqual(newgr.params['gr_source'], 3)
    
    def test_getforcenotfirst(self):
        gr = self.rebx.load_force('gr')
        gr.params['gr_source'] = 3
        cust = self.rebx.create_force('myforce')
        newgr = self.rebx.get_force('gr')
        self.assertEqual(newgr.params['gr_source'], 3)

    def test_forcenotfound(self):
        with self.assertRaises(AttributeError):
            gr = self.rebx.get_force('gr')
    
    def test_removeforce(self):
        gr = self.rebx.load_force('gr')
        self.rebx.add_force(gr)
        self.rebx.remove_force(gr)
    
    def test_removeforcenotfirst(self):
        gr = self.rebx.load_force('gr')
        self.rebx.add_force(gr)
        cust = self.rebx.create_force('myforce')
        def myforce(sim, force, particles, N):
            sim.contents.particles[1].ax += 1.e-4
        cust.update_accelerations = myforce
        cust.force_type = 'pos'
        self.rebx.add_force(cust)
        self.rebx.remove_force(gr)
    
    def test_removenonforce(self):
        with self.assertRaises(TypeError):
            self.rebx.remove_force(self.sim)

    def test_removeforcenotattached(self):
        gr = self.rebx.load_force('gr')
        with self.assertRaises(AttributeError):
            self.rebx.remove_force(gr)

class TestOperators(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1., e=0.2)
        self.rebx = reboundx.Extras(self.sim)

    def test_loadoperator(self):
        mm = self.rebx.load_operator('modify_mass')
    
    def test_loadoperatorwrongname(self):
        with self.assertRaises(RuntimeError):
            mm = self.rebx.load_force('modify_mass2')

    def test_customoperator(self):
        cust = self.rebx.create_operator('myoperator')
        def mystep(sim, operator, dt):
            sim.contents.particles[1].x += 1.e-4
        cust.step_function = mystep
        cust.operator_type = 'updater'
        self.rebx.add_operator(cust)
        self.sim.integrate(10)
        self.assertGreater(self.sim.particles[1].pomega, 1.e-3)

    def test_customopnostep(self):
        cust = self.rebx.create_operator('myoperator')
        cust.operator_type = 'updater'
        with self.assertRaises(RuntimeError):
            self.rebx.add_operator(cust)

    def test_customopnooperatortype(self):
        cust = self.rebx.create_operator('myoperator')
        def mystep(sim, operator, dt):
            sim.contents.particles[1].x += 1.e-4
        cust.step_function = mystep
        with self.assertRaises(RuntimeError):
            self.rebx.add_operator(cust)

    def test_getoperator(self):
        mm = self.rebx.load_operator('modify_mass')
        mm.params['gr_source'] = 3
        newmm = self.rebx.get_operator('modify_mass')
        self.assertEqual(newmm.params['gr_source'], 3)
    
    def test_getoperatornotfirst(self):
        mm = self.rebx.load_operator('modify_mass')
        cust = self.rebx.create_operator('myoperator')
        mm.params['gr_source'] = 3
        newmm = self.rebx.get_operator('modify_mass')
        self.assertEqual(newmm.params['gr_source'], 3)

    def test_operatornotfound(self):
        with self.assertRaises(AttributeError):
            mm = self.rebx.get_operator('modify_mass')
    
    def test_removeoperator(self):
        mm = self.rebx.load_operator('modify_mass')
        self.rebx.add_operator(mm)
        self.rebx.remove_operator(mm)

    def test_removeoperatorprepost(self):
        self.sim.integrator = "whfast"
        mm = self.rebx.load_operator('modify_mass')
        self.rebx.add_operator(mm)
        self.rebx.remove_operator(mm)
    
    def test_removemanyoperatorsteps(self):
        mm = self.rebx.load_operator('modify_mass')
        cust = self.rebx.create_operator('myoperator')
        def mystep(sim, operator, dt):
            sim.contents.particles[1].x += 1.e-4
        cust.step_function = mystep
        cust.operator_type = 'updater'
        self.rebx.add_operator(mm, dtfraction=0.5, timing='post')
        self.rebx.add_operator(cust, dtfraction=0.5, timing='post')
        self.rebx.add_operator(mm, dtfraction=0.5, timing='post')
        self.rebx.add_operator(cust, dtfraction=0.5, timing='post')
        self.rebx.add_operator(mm, dtfraction=0.5, timing='post')
        self.rebx.add_operator(cust, dtfraction=0.5, timing='pre')
        self.rebx.add_operator(mm, dtfraction=0.5, timing='pre')
        self.rebx.add_operator(cust, dtfraction=0.5, timing='pre')
        self.rebx.add_operator(mm, dtfraction=0.5, timing='pre')
        self.rebx.add_operator(cust, dtfraction=0.5, timing='pre')
        self.rebx.remove_operator(mm)
    
    def test_removenonoperator(self):
        with self.assertRaises(TypeError):
            self.rebx.remove_operator(self.sim)

    def test_removeoperatornotattached(self):
        mm = self.rebx.load_operator('modify_mass')
        with self.assertRaises(AttributeError):
            self.rebx.remove_operator(mm)

class TestAddOperator(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=1., e=0.2)
        self.rebx = reboundx.Extras(self.sim)
        self.cust = self.rebx.create_operator('addop')
        self.rebx.register_param('ctr', 'REBX_TYPE_INT')
        self.cust.params['ctr'] = 0
        def mystep(sim, operator, dt):
            operator.contents.params['ctr'] += 1
        self.cust.step_function = mystep
    
    def test_ias15updater(self):
        self.sim.integrator='ias15'
        self.cust.operator_type = 'updater'
        self.rebx.add_operator(self.cust)
        self.sim.step()
        self.assertEqual(self.cust.params['ctr'], 1)
    
    def test_ias15recorder(self):
        self.sim.integrator='ias15'
        self.cust.operator_type = 'recorder'
        self.rebx.add_operator(self.cust)
        self.sim.step()
        self.assertEqual(self.cust.params['ctr'], 1)
    
    def test_whfastupdater(self):
        self.sim.integrator='whfast'
        self.cust.operator_type = 'updater'
        self.rebx.add_operator(self.cust)
        self.sim.step()
        self.assertEqual(self.cust.params['ctr'], 2) # whfast = sec order scheme
    
    def test_whfastrecorder(self):
        self.sim.integrator='whfast'
        self.cust.operator_type = 'recorder'
        self.rebx.add_operator(self.cust)
        self.sim.step()
        self.assertEqual(self.cust.params['ctr'], 1)

    def test_mercuriusupdater(self):
        self.sim.integrator='mercurius'
        self.cust.operator_type = 'updater'
        with self.assertRaises(RuntimeError):
            self.rebx.add_operator(self.cust)
    
    def test_mercuriusrecorder(self):
        self.sim.integrator='mercurius'
        self.cust.operator_type = 'recorder'
        self.rebx.add_operator(self.cust)
        self.sim.step()
        self.assertEqual(self.cust.params['ctr'], 1)

if __name__ == '__main__':
    unittest.main()




import json
import sys
exec("import matplotlib as mpl")
exec("mpl.use(\"Agg\")")

if len(sys.argv)!=2:
    print("Usage: ipynb2py.py FILENAME")
    exit(1)
with open(sys.argv[1]) as data_file:    
    ipynb = json.load(data_file)

code = ""
for c in ipynb["cells"]:
    if c["cell_type"] == "code":
        source = c["source"]
        for s in source:
            if s[0] != "%":
                code += s.rstrip('\n')+"\n"
import socket
try:
    exec(code)
except socket.error: 
    print("A socket error occured. This is most likely due to a timeout in the NASA Horizons connections. We catch this exception here and ignore is.")
    pass



import rebound
import reboundx
import unittest
import numpy as np

"""
Acts as both a test on various integration options for forces working (add_force vs step before/after/both)
and for reproducibility with the simulationarchive under those various modes.
"""

# When the scheme error with leapfrog is larger than the perturbation, we get qualitatively wrong results, so don't test
# Cna only add effects as forces to mercurius, so only test individually with add_force
integrators = ['ias15', 'whfast']
rebxintegrators = ['euler', 'rk2', 'rk4', 'implicit_midpoint']

class TestSimulationArchive(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(m=1.e-4, a=1., e=0.2)
        self.sim.move_to_com()
        self.sim.dt = 1.e-2*self.sim.particles[1].P
        self.rebx = reboundx.Extras(self.sim)
        self.gr = self.rebx.load_force('gr')
        self.gr.params['c'] = 1e2
        self.pomega0 = self.sim.particles[1].pomega
        self.E0 = self.rebx.gr_hamiltonian(self.gr)

    def test_force(self):
        for integrator in integrators + ['mercurius']:
            for rebxintegrator in rebxintegrators:
                self.setUp()
                self.sim.integrator = integrator
                self.rebx.add_force(self.gr)
            
                self.sim.integrate(1000)
                E = self.rebx.gr_hamiltonian(self.gr)
                # Test effect is actually doing something to the pericenter
                self.assertGreater(self.sim.particles[1].pomega, 0.1, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
                # test energy conservtaion
                self.assertLess(np.abs((E-self.E0)/self.E0), 1.e-4, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
               
                # test bitwise reproducibility starting from an intermediate snapshot
                self.sim.simulationarchive_snapshot('test.sa', deletefile=True)
                self.rebx.save('test.rebx')
                self.sim.integrate(2000)
                self.sim.simulationarchive_snapshot('test.sa')

                sa = reboundx.SimulationArchive('test.sa', 'test.rebx')
                simf, rebxf = sa[-1]
                tmax = simf.t
                sim, rebx = sa[0]
                sim.integrate(tmax)
                self.assertEqual(self.sim.particles[1].x, sim.particles[1].x, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))

    def test_pre(self):
        for integrator in integrators:
            for rebxintegrator in rebxintegrators:
                self.setUp()
                self.sim.integrator = integrator
                self.sim.ri_ias15.epsilon = 0 # use fixed timesteps for ias15 if used, no problem otherwise
                self.integforce = self.rebx.load_operator("integrate_force")
                self.integforce.params['force'] = self.gr
                self.rebx.add_operator(self.integforce, dtfraction=1., timing="pre")
            
                self.sim.integrate(1000)
                E = self.rebx.gr_hamiltonian(self.gr)
                # Test effect is actually doing something to the pericenter
                self.assertGreater(self.sim.particles[1].pomega, 0.1, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
                # test energy conservtaion
                self.assertLess(np.abs((E-self.E0)/self.E0), 1.e-4, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
               
                # test bitwise reproducibility starting from an intermediate snapshot
                self.sim.simulationarchive_snapshot('test.sa', deletefile=True)
                self.rebx.save('test.rebx')
                self.sim.integrate(2000)
                self.sim.simulationarchive_snapshot('test.sa')

                sa = reboundx.SimulationArchive('test.sa', 'test.rebx')
                simf, rebxf = sa[-1]
                tmax = simf.t
                sim, rebx = sa[0]
                sim.integrate(tmax)
                self.assertEqual(self.sim.particles[1].x, sim.particles[1].x, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))

    def test_post(self):
        for integrator in integrators:
            for rebxintegrator in rebxintegrators:
                self.setUp()
                self.sim.integrator = integrator
                self.sim.ri_ias15.epsilon = 0 # use fixed timesteps for ias15 if used, no problem otherwise
                self.integforce = self.rebx.load_operator("integrate_force")
                self.integforce.params['force'] = self.gr
                self.rebx.add_operator(self.integforce, dtfraction=1., timing="post")
            
                self.sim.integrate(1000)
                E = self.rebx.gr_hamiltonian(self.gr)
                # Test effect is actually doing something to the pericenter
                self.assertGreater(self.sim.particles[1].pomega, 0.1, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
                # test energy conservtaion
                self.assertLess(np.abs((E-self.E0)/self.E0), 1.e-4, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
               
                # test bitwise reproducibility starting from an intermediate snapshot
                self.sim.simulationarchive_snapshot('test.sa', deletefile=True)
                self.rebx.save('test.rebx')
                self.sim.integrate(2000)
                self.sim.simulationarchive_snapshot('test.sa')

                sa = reboundx.SimulationArchive('test.sa', 'test.rebx')
                simf, rebxf = sa[-1]
                tmax = simf.t
                sim, rebx = sa[0]
                sim.integrate(tmax)
                self.assertEqual(self.sim.particles[1].x, sim.particles[1].x, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))

    def test_2ndorder(self):
        for integrator in integrators:
            for rebxintegrator in rebxintegrators:
                self.setUp()
                self.sim.integrator = integrator
                self.sim.ri_ias15.epsilon = 0 # use fixed timesteps for ias15 if used, no problem otherwise
                self.integforce = self.rebx.load_operator("integrate_force")
                self.integforce.params['force'] = self.gr
                self.rebx.add_operator(self.integforce)
            
                self.sim.integrate(1000)
                E = self.rebx.gr_hamiltonian(self.gr)
                # Test effect is actually doing something to the pericenter
                self.assertGreater(self.sim.particles[1].pomega, 0.1, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
                # test energy conservtaion
                self.assertLess(np.abs((E-self.E0)/self.E0), 1.e-4, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))
               
                # test bitwise reproducibility starting from an intermediate snapshot
                self.sim.simulationarchive_snapshot('test.sa', deletefile=True)
                self.rebx.save('test.rebx')
                self.sim.integrate(2000)
                self.sim.simulationarchive_snapshot('test.sa')

                sa = reboundx.SimulationArchive('test.sa', 'test.rebx')
                simf, rebxf = sa[-1]
                tmax = simf.t
                sim, rebx = sa[0]
                sim.integrate(tmax)
                self.assertEqual(self.sim.particles[1].x, sim.particles[1].x, msg='REB integrator: {0}, REBX integrator: {1}'.format(integrator, rebxintegrator))

if __name__ == '__main__':
    unittest.main()

